/*
 * Author: Affan Syed <asyed@usc.edu>
 *  
 *   simNode.cpp: Defines all node and event related functionalities of the simulator
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */

#include "simNode.h"


extern cVTime globalClock;///Our global clock!
extern vector<cNode> gNodeVector;///vector that will keep track of all the nodes in the simulation
extern Timers gTimerManager;



/*************** Event Generation *******************/

//This describes the event expiration routine. Each Event contains in itself
// the info needed to drive the simulation. 
//This should (on its own) handle all the required event generation and addition to event queue
///needed for the correct simulation

int Event::Expire(){
	//fprintf(stderr,"Timer Expires for node id %d at Global time = %d \n", eInfo_.nodeId_, globalClock.globalTime);

	vector<cNode>::iterator nodePosition;
	

	nodePosition = find(gNodeVector.begin(), gNodeVector.end(),eInfo_.nodeId_);

	if(nodePosition == gNodeVector.end()) { ///we didnt find the node id ..something fishy!
	
		cerr<<"node id "<< eInfo_.nodeId_<<" Could not be found "<<endl;
		
		delete this;//NOTE:: Weird, but this cleans up the memory!
		
		return -1;
	}

	else{ //now we have the right node execute the event
	switch (eInfo_.eType_){
		/****
		The Node/traffic layer events for generation of traffic are handled over here
		*****/
		case e_tx_end:
			///we have to handle reception of 
			if  ( (*nodePosition).NodeTransmitEnd(eInfo_) ) //this is the case when we are being asked to post a new event
				return (int)eInfo_.timeLeft_;
			else{
				
				delete this;//NOTE:: Weird, but this cleans up the memory!
				return -1;
				}
			
			break;
		case e_tx:
			if  ( (*nodePosition).NodeAppTransmit(eInfo_) ) //this is the case when we are being asked to post a new event
				return (int)eInfo_.timeLeft_;
			else{
				
				delete this;
				return -1;
				}

			break;

			/***
			MAC and Phy specific cases are handled here
			****/

		case e_mac_timeOut:
		case e_mac_slotTime://in both cases call the same function - but STLohi will itself figure out what to do 
			if  ( (*nodePosition).MAC->macTimerExpire(eInfo_) ) //this is the case when we are being asked to post a new event
				return (int)eInfo_.timeLeft_;
			else{
				
				delete this;
				return -1;
				}
			
			break;
		///first the three tone channel cases	
		case e_tone_rx_end:

			(*nodePosition).MAC->underwaterPhy.PhyToneRxEnds(eInfo_);//phyinterfaceUPDATE
			delete this;
			return -1;

			
			break;
		case e_tone_rx_start:

			(*nodePosition).MAC->underwaterPhy.PhyToneRxStart(eInfo_);//phyinterfaceUPDATE
				
			delete this;
			return -1;

			
			break;
		case e_tone_tx_end:
			(*nodePosition).MAC->underwaterPhy.PhyToneTxEnds(eInfo_);//phyinterfaceUPDATE
				
			delete this;
			return -1;

			
			break;
			
		//Now the three data channel events- all three go to both phy and data
		//the returned state of the phy layer is provided to MAC for transitional information

		case e_data_rx_end:{
			//phyinterfaceUPDATE
			channelState initialPhyState=(*nodePosition).MAC->underwaterPhy.PhyDataRxEnds(eInfo_);
			if  ( (*nodePosition).MAC->macDataRxEnd(eInfo_,initialPhyState))
				return (int)eInfo_.timeLeft_;
			else{
				
				delete this;
				return -1;
				}
			
			break;
			}
		case e_data_rx_start:{
			//phyinterfaceUPDATE
			channelState initialPhyState=(*nodePosition).MAC->underwaterPhy.PhyDataRxStart(eInfo_);
			if  ( (*nodePosition).MAC->macDataRxStart(eInfo_,initialPhyState))
				return (int)eInfo_.timeLeft_;
			else{
				
				delete this;
				return -1;
				}
			
			break;
			}
		case e_data_tx_end:{
			//phyinterfaceUPDATE
			channelState initialPhyState=(*nodePosition).MAC->underwaterPhy.PhyDataTxEnds(eInfo_);
			bool retVal=(*nodePosition).MAC->macDataTxEnd(eInfo_,initialPhyState);
			
			if  ( retVal)
				return (int)eInfo_.timeLeft_;
			else{
				
				delete this;
				return -1;
				}
			
			break;
			}
		//one purely MAC event
		case e_mac_toneDetected:
			//first make sure that the event status is unset
			 (*nodePosition).MAC->underwaterPhy.TD.isSet=false;
			if  ( (*nodePosition).MAC->macToneDetected(eInfo_))//phyinterfaceUPDATE
				return (int)eInfo_.timeLeft_;
			else{
			
				delete this;//NOTE:: Weird, but this cleans up the memory!
				return -1;
			}
			break;
		//two abstract cases which actually inform transmitters of their failed transmissions
		case e_abstract_reportLoss:

			 (*nodePosition).MAC->m_pktsLost++;
				delete this;
				return -1;			
			break;
		case e_abstract_reportLossByTone:

			 (*nodePosition).MAC->m_pktsLostByTone++;
				delete this;//NOTE:: Weird, but this cleans up the memory!
				return -1;			
			break;

		
			
		default:
			cerr<<"node id "<< eInfo_.nodeId_<<" presented with unknown event type!"<<endl;
			delete this;//NOTE:: Weird, but this cleans up the memory!
			return -1;
			break;
			
		}
	}
	
	
}

/****************************************************/


/*************** The cNode class implementation *************/
//the copy constructor and equality operator are needed b/c MAC is 'newed'

cNode::cNode(const cNode& orig){ //the copy constructor


		*(this)=orig;//I dont need to repeat separate MAC instantiation as the = operator will handle that anyway!
	
			
}
cNode& cNode::operator= (const cNode& orig){
	if (this == &orig) return *this;

			
		myId =orig.myId;// the unique node id kept for purposes of identification
		macInstantiated = orig.macInstantiated;
		
		myTraffic=orig.myTraffic;
		myMacParams= orig.myMacParams;
		hdlDeaf=orig.hdlDeaf;
		
		myClock = orig.myClock;
		myLocation = orig.myLocation;
		myState = orig.myState;
		linRegParam= orig.linRegParam;
		outFile= orig.outFile;
	


		if(macInstantiated){
			if(myMacParams.myMacType ==e_STLohi){
				
				
				//instantiate st lohi mac class
				MAC = new cSTLohi(myId,outFile,myMacParams);
				macInstantiated =true;
			}
			else if (myMacParams.myMacType ==e_UTLohi){
				//instantiate ut lohi mac class
				MAC = new cUTLohi(myId,outFile,myMacParams);
				macInstantiated =true;
			}			
			else if (myMacParams.myMacType==e_ALOHA){
				MAC = new cALOHA(myId,outFile,myMacParams);
				macInstantiated =true;				
				}
			else if (myMacParams.myMacType==e_SALOHA){
				MAC = new cSALOHA(myId,outFile,myMacParams);
				macInstantiated =true;				
				}
			else if (myMacParams.myMacType==e_TDMA){
				MAC = new cTDMA(myId,outFile,myMacParams);
				macInstantiated =true;				
				}
			else if (myMacParams.myMacType==e_CSMA){
				MAC = new cCSMA(myId,outFile,myMacParams);
				macInstantiated =true;				
				}
			//*MAC=*(orig.MAC);//why was i ever doing that!!!?
		}

		return *this;

}



void cNode::initNode(){
				
	///Do other initizalization over here
		if(myMacParams.myMacType ==e_STLohi){
				//instantiate st lohi mac class
				MAC = new cSTLohi(myId,outFile,myMacParams);
				macInstantiated =true;	
				MAC->resetMAC();
			}
			else if (myMacParams.myMacType==e_UTLohi){
				//instantiate ut lohi mac class
				MAC = new cUTLohi(myId,outFile,myMacParams);
				macInstantiated =true;				
				MAC->resetMAC();
			}
			else if (myMacParams.myMacType==e_ALOHA){
				MAC = new cALOHA(myId,outFile,myMacParams);
				macInstantiated =true;				
				MAC->resetMAC();
				}
			else if (myMacParams.myMacType==e_SALOHA){
				MAC = new cSALOHA(myId,outFile,myMacParams);
				macInstantiated =true;				
				MAC->resetMAC();
				}
			else if(myMacParams.myMacType ==e_TDMA){
				MAC = new cTDMA(myId,outFile,myMacParams);
				macInstantiated =true;				
				MAC->resetMAC();
				}
			else if(myMacParams.myMacType ==e_CSMA){
				MAC = new cCSMA(myId,outFile,myMacParams);
				macInstantiated =true;				
				MAC->resetMAC();
				}
			else {
				OUTCON(-1,"","Unknown MAC type in node instantiation",2);
				
			}
}

void cNode::resetNode(){
	///reset all values of the node for multiple simulation runs
	linRegParam.resetAll();
	
}

//This signals the data transmit end from MAC 
//In our case this should just result in another packet being generated immediately if 
//traffic type is saturate.
bool cNode::NodeTransmitEnd(EventInfo & txPkt){


//Traffic Model: Add if(check node's traffic model). and take decision based on that.
//JUST make sure it is called every time data transmission is finished

/**
 	INFINITE QUEUE CODE CHANGES	-- START
 	
if(myTraffic.eType==e_saturated){
	txPkt.eType_ = e_tx;//set up an application layer transmission event
	txPkt.nodeId_= myId;//set the transmission for this node
	
	txPkt.timeLeft_ = (timeTicks)(0); // immediately send another packet		
	int dest;
	do{//find a destination index other than me	
		dest = (int) (gNodeVector.size()* RAND_UNIFORME);
	}while(gNodeVector[dest].myId ==myId && dest <=gNodeVector.size() );//second condition gurantees bounds				

		
	txPkt.packet_.destId_ = gNodeVector[dest].myId;
	txPkt.packet_.msgType_ = e_msg_data;//obviously the application has only data pkts to send
	txPkt.packet_.sourceId_ =myId;
	if(myTraffic.ePktType ==e_constant)
		txPkt.packet_.length_ =myTraffic.pktLength;
	else if(myTraffic.ePktType==e_uniform)
		txPkt.packet_.length_ =(int)(myTraffic.pktLength*RAND_UNIFORME);
	return true;
}

***/
if(pktQueue.size()> 0){//i.e there is something in the queue
	txPkt = pktQueue.front();
	OUTCON(myId,"Popping pkt from Queue","",3);
	pktQueue.pop();//remove the member from the queue
	txPkt.timeLeft_ = (timeTicks)(RAND_UNIFORME*0.3*SEC_TO_TIMETICK);//TODO::the 0.33 is a hack right now.. please fix later
	return true;
	}
else if(myTraffic.eType==e_saturated){//OTHER WISE  check if the traffic model is saturated
	/**
 	INFINITE QUEUE CODE CHANGES	-- ends
 	***/
	txPkt.eType_ = e_tx;//set up an application layer transmission event
	txPkt.nodeId_= myId;//set the transmission for this node
	
	txPkt.timeLeft_ = (timeTicks)(0); // immediately send another packet		
	int dest;
		do{//find a destination index other than me	
			dest = (int) (gNodeVector.size()* RAND_UNIFORME);
		}while(gNodeVector[dest].myId ==myId && dest <=gNodeVector.size() );//second condition gurantees bounds				

		
	txPkt.packet_.destId_ = gNodeVector[dest].myId;
	txPkt.packet_.msgType_ = e_msg_data;//obviously the application has only data pkts to send
	txPkt.packet_.sourceId_ =myId;
	if(myTraffic.ePktType ==e_constant)
		txPkt.packet_.length_ =myTraffic.pktLength;
	else if(myTraffic.ePktType==e_uniform)
		txPkt.packet_.length_ =(int)(myTraffic.pktLength*RAND_UNIFORME);
	return true;
 	}
	return false;//otherwise no event shoudl be generated from this reception event
}

//Traffic generted - Try and send it using the MAC 
bool cNode::NodeAppTransmit(EventInfo & txPkt){

OUTCON(myId,"cNode::NodeAppTransmit","",4);
/*****
//HACK REMOVED
if(myMacParams.myMacType==e_ALOHA ||
		myMacParams.myMacType==e_SALOHA){
		if(myId==0)//the center node should never attempt transmission. 
		 return false;	
	}

****/

/**
 	INFINITE QUEUE CODE CHANGES	-- START
 
 
MAC->macSend(txPkt.packet_);//AFFAN: you can look at the return value to decide if the traffic went through, or you can 
							//ignore ARQ for now and just let the mac keep count of dropped packets. 							
***/
if(!MAC->macSend(txPkt.packet_)){//if mac is not able to send the pkt right now, put it in our local q
	OUTCON(myId,"***PUSHING**** pkt into Queue","",3);
	pktQueue.push(txPkt);
}
/***
==================================
Ignore all the rest for our toy simulation experiments. 
if you want to enable this, make sure u return true;
====================================
*/
//Traffic Model: Add if(check node's traffic model). and take decision based on that..
if(myTraffic.eType == e_CBR || 
   myTraffic.eType ==e_poisson ||
   myTraffic.eType ==e_bursty ){
	
	txPkt.eType_ = e_tx;//set up an application layer transmission event
	txPkt.nodeId_= myId;//set the transmission for this node
	if(myTraffic.eType==e_poisson)
		txPkt.timeLeft_ = (timeTicks)(random_exp(myTraffic.parameter)*SEC_TO_TIMETICK); // exponentially set up the next packet tx time
	else if (myTraffic.eType==e_CBR)
		txPkt.timeLeft_ = (timeTicks)(myTraffic.parameter*SEC_TO_TIMETICK); // CBR traffic
	else if (myTraffic.eType==e_bursty){
			if(m_burstCount < BURST_PARAM){//we are still in a bursty mode
				txPkt.timeLeft_ = (timeTicks)(random_exp(myTraffic.parameter*BURST_PARAM)
					*SEC_TO_TIMETICK); // set events at 10 times the rate of avg interarrival 
				m_burstCount++;
				}
			else{
				m_burstCount=0; 
				txPkt.timeLeft_ = (timeTicks)(random_exp(myTraffic.parameter/BURST_PARAM)
					*SEC_TO_TIMETICK); // exponentially set up the next packet tx time with 10 time the avg scale
				}

		}
		
		
	int dest;
/*** 
///HACK REMOVED 
	//HACK!!! with ALOHA test I just want to send it always to center node that must be the first addition to the node list
	//TODO:: debug and see that it does this
	if(myMacParams.myMacType==e_ALOHA ||
		myMacParams.myMacType==e_SALOHA){
		dest=0;
	}
	else{//otherwise randomly choose another destination
		do{//find a destination index other than me	
			dest = (int) (gNodeVector.size()* RAND_UNIFORME);
			}while(gNodeVector[dest].myId ==myId && dest <=gNodeVector.size() );//second condition gurantees bounds				
	}	
	***/
	do{//find a destination index other than me	
		dest = (int) (gNodeVector.size()* RAND_UNIFORME);
	}while(gNodeVector[dest].myId ==myId && dest <=gNodeVector.size() );//second condition gurantees bounds				

	txPkt.packet_.destId_ = gNodeVector[dest].myId;
	txPkt.packet_.msgType_ = e_msg_data;//obviously the application has only data pkts to send
	txPkt.packet_.sourceId_ =myId;
	if(myTraffic.ePktType ==e_constant)
		txPkt.packet_.length_ =myTraffic.pktLength;
	else if(myTraffic.ePktType==e_uniform)
		txPkt.packet_.length_ =(int)(myTraffic.pktLength*RAND_UNIFORME);
	return true;
}

return false;

}


//Return the time in the clock granularity of the local clock
//Corrected time = (a+1) localTime+b ; where a and b are
// linear regression params
timeTicks cNode::GetCorrectedTime(bool withOffset){

	timeTicks correctedTime;
	timeTicks localTime;
	double slope; //slope and offset of the linear regression line
	double offset=0;
	
	localTime= myClock.getLocalTime(globalClock.globalTime);
	slope = linRegParam.GetSlope();
		
	if(withOffset){
		offset = linRegParam.GetOffset();		
		}
	
	correctedTime =  (timeTicks)granularityConversion(((slope+1) * (double)localTime +offset));
	return correctedTime;
}

